/* Iris: micro-kernel for a capability-based operating system.
 * iris.h: C header file for userspace programs.
 * Copyright 2009 Bas Wijnen <wijnen@debian.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __IRIS_H
#define __IRIS_H

// Without the standard library, we don't have this definition.
// I preferred ((void*)0), but C++ has too strict type-checking to
// make that work.
#ifndef NULL
#define NULL 0
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define PAGE_BITS (12)
#define PAGE_SIZE (1 << PAGE_BITS)
#define PAGE_MASK (~(PAGE_SIZE - 1))

#define KERNEL_MASK 0xfff
#define CAPTYPE_MASK 0xe00
#define REQUEST_MASK (KERNEL_MASK & ~CAPTYPE_MASK)
#define CAPTYPE_RECEIVER 0x000
#define CAPTYPE_MEMORY 0x200
#define CAPTYPE_THREAD 0x400
#define CAPTYPE_PAGE 0x600
#define CAPTYPE_CAPABILITY 0x800
#define CAPTYPE_CAPPAGE 0xa00
/*#define CAPTYPE_??? 0xc00*/
/*#define CAPTYPE_??? 0xe00*/

/* This works on all kernel capabilities.  */
#define CAP_DEGRADE 0

/* Operations */
#define CAP_RECEIVER_SET_OWNER 1
#define CAP_RECEIVER_CREATE_CAPABILITY 2
#define CAP_RECEIVER_CREATE_CALL_CAPABILITY 3
#define CAP_RECEIVER_GET_REPLY_PROTECTED_DATA 4
#define CAP_RECEIVER_SET_REPLY_PROTECTED_DATA 5
#define CAP_RECEIVER_ALL_RIGHTS 0x7f
/* Not an operation; a capability with this bit set is a call capability.  */
#define CAP_RECEIVER_CALL 7
/* Same thing for reply capability.  */
#define CAP_RECEIVER_REPLY 8
/* If set on a call capability, waiting for only this reply is disabled.  */
#define CAP_RECEIVER_CALL_ASYNC 1

#define CAP_MEMORY_CREATE 1
#define CAP_MEMORY_DESTROY 2
#define CAP_MEMORY_LIST 3
#define CAP_MEMORY_MAP 4
#define CAP_MEMORY_MAPPING 5
#define CAP_MEMORY_SET_LIMIT 6
#define CAP_MEMORY_GET_LIMIT 7
#define CAP_MEMORY_DROP 8
#define CAP_MEMORY_ALL_RIGHTS 0x1ff

#define CAP_THREAD_INFO 1	/* Details of this are arch-specific.  */
#define CAP_THREAD_SCHEDULE 2
#define CAP_THREAD_DEBUG 3
#define CAP_THREAD_ALLOC_PHYSICAL 5
#define CAP_THREAD_MAKE_PRIV 6
#define CAP_THREAD_GET_TOP_MEMORY 7
#define CAP_THREAD_REGISTER_INTERRUPT 8
#define CAP_THREAD_ALL_RIGHTS 0x1f
#define CAP_THREAD_ALL_PRIV_RIGHTS (CAP_THREAD_ALL_RIGHTS | (1 << CAP_THREAD_REGISTER_INTERRUPT) | (1 << CAP_THREAD_GET_TOP_MEMORY) | (1 << CAP_THREAD_MAKE_PRIV) | (1 << CAP_THREAD_ALLOC_PHYSICAL))

/* These get/set_info are not arch-specific.  */
#define CAP_THREAD_INFO_PC ~0
#define CAP_THREAD_INFO_SP ~1
#define CAP_THREAD_INFO_FLAGS ~2
#define CAP_THREAD_INFO_SLEEP ~3
/* Flag values for processor state */
#define THREAD_FLAG_PRIV 0x80000000
#define THREAD_FLAG_WAITING 0x40000000
#define THREAD_FLAG_RUNNING 0x20000000
#define THREAD_FLAG_USER 0x1fffffff

#define CAP_PAGE_SHARE 1
#define CAP_PAGE_FLAGS 2
/* Not an operation; a capability without this bit cannot write to the page.  */
#define CAP_PAGE_WRITE 3
#define CAP_PAGE_ALL_RIGHTS 0x1ff

/* Operation details for PAGE_SHARE */
	/* Forget the source page during the operation.  This makes it a move.  */
#define PAGE_SHARE_FORGET 0x10000
	/* Make the target unwritable.  */
#define PAGE_SHARE_READONLY 0x20000
	/* Make the target independent of the source (make a copy if needed).  */
#define PAGE_SHARE_COPY 0x40000

/* Flag values for Page and Cappage objects.  */
	/* A writable page can be written to.  This flag can not be set while the frame is shared.  */
#define PAGE_FLAG_WRITABLE 1
	/* When paying, the memory's use is incremented.  If a frame is held, it cannot be lost.  Frames are lost when the last payer forgets them.  */
#define PAGE_FLAG_PAYING 2
	/* Set if this page has a frame associated with it.  This flag is automatically reset if the frame is lost because of payment problems.  */
#define PAGE_FLAG_FRAME 4
	/* This is a read-only flag, which is set if the Page is shared.  */
#define PAGE_FLAG_SHARED 8
	/* This is a read-only flag, saying if this is physical memory, which mustn't be freed.  */
#define PAGE_FLAG_PHYSICAL 0x10
	/* This is a read-only flag, saying if this is uncachable memory.  */
#define PAGE_FLAG_UNCACHED 0x20

#define CAP_CAPABILITY_GET 1
#define CAP_CAPABILITY_ALL_RIGHTS 0x1ff

#define CAPPAGE_SIZE 102
/* Cappage has Page's operations as well.  */
#define CAP_CAPPAGE_SET 4
#define CAP_CAPPAGE_ALL_RIGHTS 0x1ff

#ifndef __KERNEL
typedef unsigned Capability;

extern Capability __my_receiver;
extern Capability __my_thread;
extern Capability __my_memory;
extern Capability __my_call;

Capability cap_copy (Capability src)
{
	return src | 2;
}

typedef struct Message
{
	unsigned data[4];
	Capability cap[4];
	unsigned protected_data; /* only used for receiving.  */
} Message;

static int invoke (Capability target, Message *msg)
{
	unsigned ret;
	__asm__ volatile ("lw $v0, %1\n"
			"\tlw $a3, %2\n"
			"\tlw $t0, 0($a3)\n"
			"\tlw $t1, 4($a3)\n"
			"\tlw $t2, 8($a3)\n"
			"\tlw $t3, 12($a3)\n"
			"\tlw $a0, 16($a3)\n"
			"\tlw $a1, 20($a3)\n"
			"\tlw $a2, 24($a3)\n"
			"\tlw $a3, 28($a3)\n"
			"\tsyscall\n"
			"\tmove %0, $v0"
			: "=r"(ret)
			: "m"(target), "m"(msg)
			: "v0", "t0", "t1", "t2", "t3", "a0", "a1", "a2", "a3");
	return ret;
}

static int wait (Message *msg)
{
	int ret;
	__asm__ volatile ("li $v0, 2\n"
			"\tsyscall\n"
			"\tsw $v0, %0\n"
			"\tlw $v0, %1\n"
			"\tsw $t0, 0($v0)\n"
			"\tsw $t1, 4($v0)\n"
			"\tsw $t2, 8($v0)\n"
			"\tsw $t3, 12($v0)\n"
			"\tsw $a0, 16($v0)\n"
			"\tsw $a1, 20($v0)\n"
			"\tsw $a2, 24($v0)\n"
			"\tsw $a3, 28($v0)\n"
			"\tsw $v1, 32($v0)"
			: "=m"(ret)
			: "m"(msg)
			: "memory", "v0", "v1", "t0", "t1", "t2", "t3", "a0", "a1", "a2", "a3");
	return ret;
}

static int call (Capability target, Message *msg)
{
	unsigned ret;
	Capability t = cap_copy (target);
	__asm__ volatile ("lw $v0, %1\n"
			"\tlw $v1, %2\n"
			"\tlw $t0, 0($v1)\n"
			"\tlw $t1, 4($v1)\n"
			"\tlw $t2, 8($v1)\n"
			"\tlw $t3, 12($v1)\n"
			"\tlw $a0, 16($v1)\n"
			"\tlw $a1, 20($v1)\n"
			"\tlw $a2, 24($v1)\n"
			"\tlw $a3, 28($v1)\n"
			"\tsyscall\n"
			"\tsw $v0, %0\n"
			"\tlw $v0, %2\n"
			"\tsw $t0, 0($v0)\n"
			"\tsw $t1, 4($v0)\n"
			"\tsw $t2, 8($v0)\n"
			"\tsw $t3, 12($v0)\n"
			"\tsw $a0, 16($v0)\n"
			"\tsw $a1, 20($v0)\n"
			"\tsw $a2, 24($v0)\n"
			"\tsw $a3, 28($v0)\n"
			"\tsw $v1, 32($v0)"
			: "=m"(ret)
			: "m"(t), "m"(msg)
			: "memory", "v0", "v1", "t0", "t1", "t2", "t3", "a0", "a1", "a2", "a3");
}

static int invoke_01 (Capability t, unsigned d)
{
	Message msg;
	int ret;
	msg.data[0] = d;
	return invoke (t, &msg);
}

static int invoke_02 (Capability t, unsigned d0, unsigned d1)
{
	Message msg;
	int ret;
	msg.data[0] = d0;
	msg.data[1] = d1;
	return invoke (t, &msg);
}

static int invoke_04 (Capability t, unsigned d0, unsigned d1, unsigned d2, unsigned d3)
{
	Message msg;
	int ret;
	msg.data[0] = d0;
	msg.data[1] = d1;
	msg.data[2] = d2;
	msg.data[3] = d3;
	return invoke (t, &msg);
}

static int invoke_11 (Capability t, Capability c, unsigned d)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d;
	return invoke (t, &msg);
}

static int invoke_12 (Capability t, Capability c, unsigned d0, unsigned d1)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d0;
	msg.data[1] = d1;
	return invoke (t, &msg);
}

static Capability call_c01 (Capability c, unsigned d)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d;
	ret = call (__my_call, &msg);
	return ret ? msg.cap[0] : 0;
}

static Capability call_c02 (Capability c, unsigned d0, unsigned d1)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d0;
	msg.data[1] = d1;
	ret = call (__my_call, &msg);
	return ret ? msg.cap[0] : 0;
}

static Capability call_c12 (Capability c, Capability c1, unsigned d0, unsigned d1)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.cap[1] = c1;
	msg.data[0] = d0;
	msg.data[1] = d1;
	ret = call (__my_call, &msg);
	return ret ? msg.cap[0] : 0;
}

static unsigned call_n01 (Capability c, unsigned d)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d;
	ret = call (__my_call, &msg);
	return ret ? msg.data[0] : 0;
}

static unsigned call_n02 (Capability c, unsigned d0, unsigned d1)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d0;
	msg.data[1] = d1;
	ret = call (__my_call, &msg);
	return ret ? msg.data[0] : 0;
}

static unsigned call_n03 (Capability c, unsigned d0, unsigned d1, unsigned d2)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d0;
	msg.data[1] = d1;
	msg.data[2] = d2;
	ret = call (__my_call, &msg);
	return ret ? msg.data[0] : 0;
}

static unsigned call_n04 (Capability c, unsigned d0, unsigned d1, unsigned d2, unsigned d3)
{
	Message msg;
	int ret;
	msg.cap[0] = c;
	msg.data[0] = d0;
	msg.data[1] = d1;
	msg.data[2] = d2;
	msg.data[3] = d3;
	ret = call (__my_call, &msg);
	return ret ? msg.data[0] : 0;
}

static Capability degrade (Capability src, unsigned mask)
{
	return call_c02 (src, CAP_DEGRADE, mask);
}

static void kdebug (unsigned code)
{
	invoke_02 (__my_thread, CAP_THREAD_DEBUG, code);
}

static void schedule ()
{
	invoke_01 (__my_thread, CAP_THREAD_SCHEDULE);
}

static void register_interrupt (unsigned num)
{
	invoke_12 (__my_thread, __my_receiver, CAP_THREAD_REGISTER_INTERRUPT, num);
}

static Capability get_top_memory ()
{
	return call_c01 (__my_thread, CAP_THREAD_GET_TOP_MEMORY);
}

static void unregister_interrupt (unsigned num)
{
	invoke_02 (__my_thread, CAP_THREAD_REGISTER_INTERRUPT, num);
}

static void alloc_physical (Capability page, unsigned address, int cachable)
{
	invoke_12 (__my_thread, page, CAP_THREAD_ALLOC_PHYSICAL, (address & PAGE_MASK) | (cachable ? 1 : 0));
}

static int receiver_set_owner (Capability receiver, Capability owner)
{
	return invoke_11 (receiver, owner, CAP_RECEIVER_SET_OWNER);
}

static Capability receiver_create_capability (Capability receiver, unsigned protected_data)
{
	return call_c02 (receiver, CAP_RECEIVER_CREATE_CAPABILITY, protected_data);
}

static int receiver_get_reply_protected_data (Capability receiver, unsigned data)
{
	return call_n01 (receiver, CAP_RECEIVER_GET_REPLY_PROTECTED_DATA);
}

static int receiver_set_reply_protected_data (Capability receiver, unsigned data)
{
	return invoke_02 (receiver, CAP_RECEIVER_SET_REPLY_PROTECTED_DATA, data);
}

static Capability receiver_create_call_capability (Capability receiver)
{
	return call_c02 (receiver, CAP_RECEIVER_CREATE_CALL_CAPABILITY, 0);
}

static Capability receiver_create_async_call_capability (Capability receiver)
{
	return call_c02 (receiver, CAP_RECEIVER_CREATE_CALL_CAPABILITY, 1);
}

static Capability memory_create (Capability memory, unsigned type)
{
	return call_c02 (memory, CAP_MEMORY_CREATE, type);
}

static Capability memory_create_page (Capability memory)
{
	return memory_create (memory, CAPTYPE_PAGE | REQUEST_MASK);
}

static Capability memory_create_thread (Capability memory)
{
	return memory_create (memory, CAPTYPE_THREAD | REQUEST_MASK);
}

static Capability memory_create_receiver (Capability memory)
{
	return memory_create (memory, CAPTYPE_RECEIVER | REQUEST_MASK);
}

static Capability memory_create_memory (Capability memory)
{
	return memory_create (memory, CAPTYPE_MEMORY | REQUEST_MASK);
}

static Capability memory_create_cappage (Capability memory)
{
	return memory_create (memory, CAPTYPE_CAPPAGE | REQUEST_MASK);
}

static int memory_destroy (Capability memory, Capability target)
{
	return invoke_11 (memory, target, CAP_MEMORY_DESTROY);
}

/* TODO: #define CAP_MEMORY_LIST 3 */

static int memory_map (Capability memory, Capability page, unsigned address, int writable)
{
	if (writable)
		address |= 1 << CAP_PAGE_WRITE;
	return invoke_12 (memory, page, CAP_MEMORY_MAP, address);
}

static Capability memory_mapping (Capability memory, unsigned address)
{
	return call_c02 (memory, CAP_MEMORY_MAPPING, address);
}

static void drop (Capability cap)
{
	invoke_11 (__my_memory, cap, CAP_MEMORY_DROP);
}

static Capability thread_make_priv (Capability thread)
{
	return call_c12 (__my_thread, thread, CAP_THREAD_MAKE_PRIV, ~0);
}

static unsigned thread_info (Capability thread, unsigned info, unsigned value, unsigned mask)
{
	return call_n04 (thread, CAP_THREAD_INFO, info, value, mask);
}

static unsigned thread_set_pc (Capability thread, unsigned pc)
{
	return thread_info (thread, CAP_THREAD_INFO_PC, pc, ~0);
}

static unsigned thread_set_sp (Capability thread, unsigned sp)
{
	return thread_info (thread, CAP_THREAD_INFO_SP, sp, ~0);
}

static unsigned thread_flags (Capability thread, unsigned value, unsigned mask)
{
	return thread_info (thread, CAP_THREAD_INFO_FLAGS, value, mask);
}

static unsigned thread_run (Capability thread, int run)
{
	return thread_flags (thread, run ? THREAD_FLAG_RUNNING : 0, THREAD_FLAG_RUNNING);
}

static unsigned thread_wait (Capability thread, int wait)
{
	return thread_flags (thread, wait ? THREAD_FLAG_WAITING : 0, THREAD_FLAG_WAITING);
}

static unsigned thread_sleep (Capability thread, unsigned value)
{
	return thread_info (thread, CAP_THREAD_INFO_SLEEP, value, ~0);
}

static int my_sleep (unsigned value, Message *ret)
{
	ret->data[0] = CAP_THREAD_INFO;
	ret->data[1] = CAP_THREAD_INFO_SLEEP;
	ret->data[2] = value;
	ret->data[3] = ~0;
	ret->cap[0] = 0;
	ret->cap[1] = 0;
	ret->cap[2] = 0;
	ret->cap[3] = 0;
	return call (__my_thread, ret);
}

static unsigned thread_get_sleep (Capability thread)
{
	return thread_info (thread, CAP_THREAD_INFO_SLEEP, 0, 0);
}

static unsigned thread_get_pc (Capability thread)
{
	return thread_info (thread, CAP_THREAD_INFO_PC, 0, 0);
}

static unsigned thread_get_sp (Capability thread)
{
	return thread_info (thread, CAP_THREAD_INFO_SP, 0, 0);
}

static int page_share (Capability page, Capability target, unsigned flags)
{
	return invoke_12 (page, target, CAP_PAGE_SHARE, flags);
}

static unsigned page_flags (Capability page, unsigned new_flags, unsigned mask)
{
	return call_n03 (page, CAP_PAGE_FLAGS, new_flags, mask);
}

static Capability capability_get (Capability cap)
{
	return call_c01 (cap, CAP_CAPABILITY_GET);
}

static int cappage_set (Capability page, Capability cap, unsigned index)
{
	return invoke_12 (page, cap, CAP_CAPPAGE_SET, index);
}

#endif

#ifdef __cplusplus
}
#endif

#endif
